#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Converts a Yacc/Bison grammar definition into a Python skeleton that uses ptk.
"""

import io
import getopt
import sys
import collections
import codecs
import re

from ptk.parser import production, LRParser, ParseError
from ptk.lexer import token, ReLexer, EOF
from ptk.regex import buildRegex, DeadState


Symbol = collections.namedtuple('Symbol', ('name', 'argname'))


class Options(object):
    def __init__(self, opts):
        self.compact = False
        self.arguments = False
        self.filename = None
        for opt, val in opts:
            if opt in ('-c', '--compact'):
                self.compact = True
            if opt in ('-a', '--arguments'):
                self.arguments = True
            if opt in ('-o', '--output'):
                self.filename = val
            if opt in ('-h', '--help'):
                self.usage()

        if self.compact and self.arguments:
            print('--compact and --arguments are not compatible')
            self.usage(1)

        if self.filename is None:
            print('Output file not specified')
            self.usage(1)

    def usage(self, exitCode=0):
        print('Usage: %s [options] filename' % sys.argv[0])
        print('Options:')
        print('  -h, --help      Print this')
        print('  -c, --compact   Create one method for all alternatives of a production')
        print('  -o, --output <filename> Output to file (mandatory)')
        print('  -a, --arguments Generate argument names for items in productions (incompatible with --compact)')
        sys.exit(exitCode)

    @staticmethod
    def create():
        opts, args = getopt.getopt(sys.argv[1:], 'caho:', ['compact', 'arguments', 'help', 'output='])
        return Options(opts), args


class NullToken(object):
    def __init__(self, endMarker):
        self.__rx = buildRegex('(.|\n)*%s' % re.escape(endMarker)).start()

    def feed(self, char):
        try:
            if self.__rx.feed(char):
                return None, None
        except DeadState:
            return None, None


class YaccParser(LRParser, ReLexer):
    def __init__(self, options, stream):
        self.stream = stream
        self.options = options
        super().__init__()

        self.state = 0
        self.yaccStartSymbol = None
        self.allTokens = list()
        self.allProductions = list()
        self.precedences = list()

    # Lexer

    @token(r'%\{', types=[])
    def c_decl(self, tok):
        self.set_consumer(NullToken('%}'))

    @token(r'/\*', types=[])
    def comment(self, tok):
        self.set_consumer(NullToken('*/'))

    @token(r'%union\s*{', types=[]) # Hum, no LF possible before {
    def union(self, tok):
        self.set_consumer(NullToken('}'))

    @token(r'%%')
    def part_sep(self, tok):
        self.state += 1
        if self.state == 2:
            # Ignore C code after last %%
            class IgnoreCCode(object):
                def feed(self, char):
                    if char is EOF:
                        return EOF, EOF
            self.set_consumer(IgnoreCCode())

    @staticmethod
    def ignore(char):
        return char in [' ', '\t', '\n']

    @token(r'%(left|right|nonassoc)')
    def assoc_decl(self, tok):
        pass

    @token(r'[a-zA-Z_][a-zA-Z0-9_]*')
    def identifier(self, tok):
        pass

    @token('[1-9][0-9]*')
    def number(self, tok):
        tok.value = int(tok.value)

    @token('"')
    def string(self, tok):
        class StringParser(object):
            def __init__(self):
                self.state = 0
                self.value = io.StringIO()
            def feed(self, char):
                if self.state == 0:
                    if char == '"':
                        return 'string', self.value.getvalue()
                    if char == '\\':
                        self.state = 1
                    else:
                        self.value.write(char)
                elif self.state == 1:
                    self.value.write(char)
                    self.state = 0
        self.set_consumer(StringParser())

    @token(r'\{')
    def semantic_action(self, tok):
        # Don't try to be too smart; just balance {} that are not in string litterals
        class CSemanticAction(object):
            def __init__(self):
                self.state = 0
                self.count = 1
                self.value = io.StringIO()
                self.value.write('{')

            def feed(self, char):
                self.value.write(char)
                if self.state == 0: # Nothing special
                    if char == '}':
                        self.count -= 1
                        if self.count == 0:
                            return 'semantic_action', self.value.getvalue()
                    elif char == '{':
                        self.count += 1
                    elif char == '\\':
                        self.state = 1
                    elif char == '\'':
                        self.state = 2
                    elif char == '"':
                        self.state = 4
                elif self.state == 1: # Escaping single char
                    self.state = 0
                elif self.state == 2: # Character litteral. Not that this accepts several characters
                    if char == '\\':
                        self.state = 3
                    elif char == '\'':
                        self.state = 0
                elif self.state == 3: # Escaping in character litteral
                    self.state = 2
                elif self.state == 4: # In string litteral
                    if char == '\\':
                        self.state = 5
                    elif char == '"':
                        self.state = 0
                elif self.state == 5: # Escaping in string litteral
                    self.state = 4
        self.set_consumer(CSemanticAction())

    @token(r'\'.\'')
    def litteral_token(self, tok):
        tok.value = tok.value[1]

    # Parser

    @production('YACC_FILE -> META_DECLARATION* part_sep PRODUCTION_DECL*')
    def yacc_file(self):
        pass

    # Tokens, start symbol, etc

    @production('META_DECLARATION -> "%token" identifier+<tokens>')
    def token_declaration(self, tokens):
        self.allTokens.extend(tokens)

    @production('META_DECLARATION -> assoc_decl<assoc> identifier+<tokens>')
    def assoc_declaration(self, assoc, tokens):
        self.precedences.append((assoc, tokens))

    @production('META_DECLARATION -> "%start" identifier<name>')
    def start_declaration(self, name):
        self.yaccStartSymbol = name

    @production('META_DECLARATION -> "%type" identifier identifier+')
    @production('META_DECLARATION -> "%expect" number')
    @production('META_DECLARATION -> "%debug"')
    @production('META_DECLARATION -> "%defines"')
    @production('META_DECLARATION -> "%destructor" semantic_action identifier+')
    @production('META_DECLARATION -> "%file-prefix" "=" string')
    @production('META_DECLARATION -> "%locations"')
    @production('META_DECLARATION -> "%name-prefix" "=" string')
    @production('META_DECLARATION -> "%no-parser')
    @production('META_DECLARATION -> "%no-lines')
    @production('META_DECLARATION -> "%output" "=" string')
    @production('META_DECLARATION -> "%pure-parser"')
    @production('META_DECLARATION -> "%token-table"')
    @production('META_DECLARATION -> "%verbose"')
    @production('META_DECLARATION -> "%yacc"')
    def ignored_declaration(self):
        pass

    # Productions

    @production('PRODUCTION_DECL -> identifier<left> ":" PRODUCTION_RIGHT+("|")<right> ";"')
    def production_decl(self, left, right):
        self.allProductions.append((left, right))

    @production('PRODUCTION_RIGHT -> SYMBOL*<symbols>')
    def production_right(self, symbols):
        names = list()
        indexes = dict()
        for symbol in symbols:
            if symbol.argname is None:
                names.append((symbol.name, None))
            else:
                index = indexes.get(symbol.argname, 0)
                argname = symbol.argname if index == 0 else '%s_%d' % (symbol.argname, index + 1)
                indexes[symbol.argname] = index + 1
                names.append((symbol.name, argname))

        return dict(names=names, action=None, precedence=None)

    @production('PRODUCTION_RIGHT -> PRODUCTION_RIGHT<prod> semantic_action<action>')
    def production_right_action(self, prod, action):
        if prod['action'] is not None:
            raise RuntimeError('Duplicate semantic action "%s"' % action)
        prod['action'] = action
        return prod

    @production('PRODUCTION_RIGHT -> PRODUCTION_RIGHT<prod> "%prec" identifier<prec>')
    def production_right_prec(self, prod, prec):
        if prod['precedence'] is not None:
            raise RuntimeError('Duplicate precedence declaration "%s"' % prec)
        prod['precedence'] = prec
        return prod

    @production('SYMBOL -> identifier<tok>')
    def symbol_from_identifier(self, tok):
        return Symbol(tok, None if tok in self.allTokens else tok)

    @production('SYMBOL -> litteral_token<tok>')
    def symbol_from_litteral(self, tok):
        return Symbol('"%s"' % tok, None)

    def on_start_symbol(self, result):
        self.stream.write('from ptk.lexer import ReLexer, token\n')
        self.stream.write('from ptk.parser import LRParser, production, left_assoc, right_assoc, non_assoc\n')
        self.stream.write('\n')

        for assocType, tokens in self.precedences:
            self.stream.write('@%s(%s)\n' % ({'%left': 'left_assoc', '%right': 'right_assoc', '%nonassoc': 'non_assoc'}[assocType],
                                             ', '.join([repr(tok) for tok in tokens])))
        self.stream.write('class Parser(LRParser, ReLexer):\n')
        if self.yaccStartSymbol is not None:
            self.stream.write('    startSymbol = %s\n' % repr(self.yaccStartSymbol))
            self.stream.write('\n')

        self.stream.write('    # Lexer\n')
        for name in self.allTokens:
            self.stream.write('\n')
            self.stream.write('    @token(r\'\')\n')
            self.stream.write('    def %s(self, tok):\n' % name)
            self.stream.write('        pass\n')

        methodIndexes = dict()
        def methodName(name):
            index = methodIndexes.get(name, 0)
            methodIndexes[name] = index + 1
            return name if index == 0 else '%s_%d' % (name, index + 1)

        for name, prods in self.allProductions:
            for prod in prods:
                if not self.options.compact:
                    self.stream.write('\n')
                if prod['action'] is not None:
                    for line in prod['action'].split('\n'):
                        self.stream.write('    # %s\n' % line)
                symnames = []
                for aname, argname in prod['names']:
                    symnames.append(aname if argname is None or not self.options.arguments else '%s<%s>' % (aname, argname))
                self.stream.write('    @production(\'%s -> %s\'' % (name, ' '.join(symnames)))
                if prod['precedence'] is not None:
                    self.stream.write(', priority=%s' % repr(prod['precedence']))
                self.stream.write(')\n')
                if not self.options.compact:
                    self.stream.write('    def %s(self' % methodName(name))
                    if self.options.arguments:
                        for aname, argname in prod['names']:
                            if argname is not None:
                                self.stream.write(', %s' % argname)
                    self.stream.write('):\n')
                    self.stream.write('        pass\n')
            if self.options.compact:
                self.stream.write('    def %s(self):\n' % methodName(name))
                self.stream.write('        pass\n')
                self.stream.write('\n')


def main(filename):
    import time
    options, filenames = Options.create()
    for filename in filenames:
        with codecs.getreader('utf_8')(open(filename, 'rb')) as fileobj:
            output = sys.stdout if options.filename == '-' else codecs.getwriter('utf_8')(open(options.filename, 'wb'))
            parser = YaccParser(options, output)
            t0 = time.time()
            try:
                parser.parse(fileobj.read())
            except ParseError as exc:
                print('Parse error: %s' % exc)
                tokens = exc.expecting()
                if tokens:
                    print('Was expecting %s' % ', '.join(map(repr, sorted(tokens))))
                sys.exit(1)
            finally:
                print('== Parsed file in %d ms.' % int(1000 * (time.time() - t0)))


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=logging.WARNING, format='%(asctime)-15s %(levelname)-8s %(name)-15s %(message)s')

    import sys
    main(sys.argv[1])



# -*- coding: utf-8 -*-

# (c) Jérôme Laheurte 2015-2026
# See LICENSE.txt

import functools
import collections
import logging
import re
import warnings

from ptk.lexer import EOF, token, LexerPosition, ReLexer
from ptk.grammar import Grammar, Production, GrammarError
# production is only imported so that client code doesn't have to import it from grammar
from ptk.grammar import production # pylint: disable=W0611
from ptk.utils import Singleton, callback_by_name


class ParseError(Exception):
    """
    Syntax error when parsing.

    :ivar token: The unexpected token.
    """
    def __init__(self, grammar, tok, state, tokens):
        self.token = tok
        super().__init__(f'Unexpected token "{tok.value}" ({tok.type}) in state "{sorted(state)}"')

        self._state = state
        self._expecting = set()
        for terminal in grammar.token_types():
            if grammar.__actions__.get((state, terminal), None) is not None:
                self._expecting.add(terminal)
        self._tokens = tokens

    @property
    def position(self):
        return self.token.position

    def expecting(self):
        """
        Returns a set of tokens types that would have been valid in input.
        """
        return self._expecting

    def state(self):
        """
        Returns the parser state when the error was encountered.
        """
        return self._state

    def last_token(self):
        """
        Returns the last valid token seen before this error
        """
        return self._tokens[-1]

    def lastToken(self): # pylint: disable=invalid-name
        warnings.warn('lastToken is deprecated in favor of last_token', DeprecationWarning)
        return self.last_token()

    def tokens(self):
        """
        Returns all tokens seen
        """
        return self._tokens


def left_assoc(*operators):
    """
    Class decorator for left associative operators. Use this to
    decorate your :py:class:`Parser` class. Operators passed as
    argument are assumed to have the same priority. The later you
    declare associativity, the higher the priority; so the following
    code

    .. code-block:: python

       @left_assoc('+', '-')
       @left_assoc('*', '/')
       class MyParser(LRParser):
           # ...

    declares '+' and '-' to be left associative, with the same
    priority. '*' and '/' are also left associative, with a higher
    priority than '+' and '-'.

    See also the *priority* argument to :py:func:`production`.
    """
    def _wrapper(cls):
        cls.__precedence__.insert(0, ('left', set(operators)))
        return cls
    return _wrapper


def leftAssoc(*operators): # pylint: disable=invalid-name
    warnings.warn('leftAssoc is deprecated in favor of left_assoc', DeprecationWarning)
    return left_assoc(*operators)


def right_assoc(*operators):
    """
    Class decorator for right associative operators. Same remarks as :py:func:`left_assoc`.
    """
    def _wrapper(cls):
        cls.__precedence__.insert(0, ('right', set(operators)))
        return cls
    return _wrapper


def rightAssoc(*operators): # pylint: disable=invalid-name
    warnings.warn('rightAssoc is deprecated in favor of right_assoc', DeprecationWarning)
    return right_assoc(*operators)


def non_assoc(*operators):
    """
    Class decorator for non associative operators. Same remarks as :py:func:`left_assoc`.
    """
    def _wrapper(cls):
        cls.__precedence__.insert(0, ('non', set(operators)))
        return cls
    return _wrapper


def nonAssoc(*operators): # pylint: disable=invalid-name
    warnings.warn('nonAssoc is deprecated in favor of non_assoc', DeprecationWarning)
    return non_assoc(*operators)


class _StartState(metaclass=Singleton): # pylint: disable=too-few-public-methods
    __reprval__ = '\u03A3'


class _ResolveError(Exception):
    pass


@functools.total_ordering
class _Item:
    __slots__ = ('production', 'dot', 'terminal', 'index', 'should_reduce', 'expecting')

    def __init__(self, prod, dot, terminal):
        self.production = prod
        self.dot = dot
        self.terminal = terminal
        self.index = None
        self.should_reduce = self.dot == len(self.production.right)
        self.expecting = None if self.should_reduce else self.production.right[self.dot]

    def next(self):
        """
        Returns an item with the dot advanced one position
        """
        return _Item(self.production, self.dot + 1, self.terminal)

    def __repr__(self):
        symbols = list(self.production.right)
        symbols.insert(self.dot, '\u2022')
        return '%s -> %s (%s)' % (self.production.name, ' '.join([repr(sym) for sym in symbols]), self.terminal) # pylint: disable=consider-using-f-string

    def __eq__(self, other):
        return (self.production, self.dot, self.terminal) == (other.production, other.dot, other.terminal)

    def __lt__(self, other):
        return (self.production, self.dot, self.terminal) < (other.production, other.dot, other.terminal)

    def __hash__(self):
        return hash((self.production, self.dot, self.terminal))


class _Accept(BaseException):
    def __init__(self, result):
        self.result = result
        super().__init__()


_StackItem = collections.namedtuple('_StackItem', ['state', 'value', 'position'])


class _Shift: # pylint: disable=too-few-public-methods
    def __init__(self, new_state):
        self.new_state = new_state

    def do_action(self, grammar, stack, tok): # pylint: disable=W0613
        stack.append(_StackItem(self.new_state, tok.value, tok.position))
        return True


class _Reduce: # pylint: disable=too-few-public-methods
    def __init__(self, item):
        self.item = item
        self.nargs = len(item.production.right)

    def do_action(self, grammar, stack, tok): # pylint: disable=W0613
        pos, (callback, kwargs) = self._get_callback(stack)
        self._applied(grammar, stack, callback(grammar, **kwargs), pos)
        return False

    def _applied(self, grammar, stack, prod_val, position):
        stack.append(_StackItem(grammar.goto(stack[-1].state, self.item.production.name), prod_val, position))

    def _get_callback(self, stack):
        if self.nargs:
            args = [stackItem.value for stackItem in stack[-self.nargs:]]
            pos = stack[-self.nargs].position
            stack[-self.nargs:] = []
        else:
            args = []
            pos = stack[-1].position # Hum.
        return pos, self.item.production.apply(args, pos)


class LRParser(Grammar):
    """
    LR(1) parser. This class is intended to be used with a lexer class
    derived from :py:class:`LexerBase`, using inheritance; it
    overrides :py:func:`LexerBase.new_token` so you must inherit from
    the parser first, then the lexer:

    .. code-block:: python

       class MyParser(LRParser, ReLexer):
           # ...

    """
    def __init__(self): # pylint: disable=R0914,R0912
        super().__init__()
        self._restart_parser()

    def rstack(self):
        return reversed(self.__stack)

    def new_token(self, tok):
        try:
            for action, stack in self._process_token(tok):
                if action.do_action(self, stack, tok):
                    break
            self.__tokens.append(tok)
            return None
        except _Accept as exc:
            self._restart_parser()
            self.on_start_symbol(exc.result)
            return exc.result

    def _process_token(self, tok):
        while True:
            action = self.__actions__.get((self.__stack[-1].state, tok.type), None)
            if action is None:
                raise ParseError(self, tok, self.__stack[-1].state, self.__tokens)
            yield action, self.__stack

    def on_start_symbol(self, start_symbol): # pragma: no cover
        """
        This is called when the start symbol has been reduced.

        :param start_symbol: The value associated with the start symbol.
        """
        return self.newSentence(start_symbol)

    def newSentence(self, start_symbol): # pylint: disable=invalid-name,unused-argument
        warnings.warn('newSentence has been deprecated in favor of on_start_symbol', DeprecationWarning)
        raise NotImplementedError

    @classmethod
    def _create_production_parser(cls, name, priority, attrs):
        return ProductionParser(callback_by_name(name), priority, cls, attrs)

    @classmethod
    def _create_reduce_action(cls, item):
        return _Reduce(item)

    @classmethod
    def _create_shift_action(cls, state):
        return _Shift(state)

    @classmethod
    def prepare(cls):
        for prod in cls.productions():
            if prod.name is _StartState:
                break
        else:
            def acceptor(_, result):
                raise _Accept(result)
            prod = Production(_StartState, acceptor)
            prod.add_symbol(cls._default_start_symbol() if cls.startSymbol is None else cls.startSymbol, name='result')
            cls.__productions__.insert(0, prod)

        cls.startSymbol = _StartState
        super().prepare()

        states, goto = cls.__compute_states(prod)
        reachable = cls.__compute_actions(states, goto)

        logger = logging.getLogger('LRParser')
        cls.__resolve_conflicts(logger)

        used_tokens = {symbol for _, symbol in cls.__actions__ if symbol is not EOF}
        if used_tokens != cls.token_types(): # pragma: no cover pylint: disable=no-member
            logger.warning('The following tokens are not used: %s', ','.join([repr(sym) for sym in sorted(cls.token_types() - used_tokens)])) # pylint: disable=no-member

        if reachable != cls.nonterminals(): # pragma: no cover
            logger.warning('The following nonterminals are not reachable: %s', ','.join([repr(sym) for sym in sorted(cls.nonterminals() - reachable)]))

        # Reductions only need goto entries for nonterminals
        cls._goto = {(state, symbol): new_state for (state, symbol), new_state in goto.items() if symbol not in cls.token_types()} # pylint: disable=no-member

        parts = []
        if cls.nSR:
            parts.append(f'{cls.nSR} shift/reduce conflicts')
        if cls.nRR:
            parts.append(f'{cls.nRR} reduce/reduce conflicts')
        if parts:
            logger.warning(', '.join(parts))

        # Cast to tuple because sets are not totally ordered
        for index, state in enumerate([tuple(cls._startState)] + sorted([tuple(state) for state in states if state != cls._startState])):
            logger.debug('State %d', index)
            for item in sorted(state):
                logger.debug('    %s', item)
                item.index = index
            cls.__lrstates__.append(sorted(state)) # pylint: disable=no-member
        logger.info('%d states.', len(states))

    @classmethod
    def __compute_states(cls, start):
        all_syms = list(cls.token_types() | cls.nonterminals()) # pylint: disable=no-member
        goto = []
        cls._startState = frozenset([_Item(start, 0, EOF)])
        states = set([cls._startState])
        stack = [cls._startState]
        while stack:
            state = stack.pop()
            state_closure = cls.__item_set_closure(state)
            for symbol in all_syms:
                # Compute goto(symbol, state)
                next_state = frozenset([item.next() for item in state_closure if item.expecting == symbol])
                if next_state:
                    goto.append(((state, symbol), next_state))
                    if next_state not in states:
                        states.add(next_state)
                        stack.append(next_state)
        return states, dict(goto)

    @classmethod
    def __compute_actions(cls, states, goto):
        cls.__actions__ = {}
        reachable = set()
        token_types = cls.token_types() # pylint: disable=no-member
        for state in states:
            for item in cls.__item_set_closure(state):
                if item.should_reduce:
                    action = cls._create_reduce_action(item)
                    reachable.add(item.production.name)
                    cls.__add_reduce_action(state, item.terminal, action)
                else:
                    symbol = item.production.right[item.dot]
                    if symbol in token_types:
                        cls.__add_shift_action(state, symbol, cls._create_shift_action(goto[(state, symbol)]))
        return reachable

    @classmethod
    def __should_prefer_shift(cls, logger, reduce_action, symbol):
        logger.info('Shift/reduce conflict for "%s" on "%s"', reduce_action.item, symbol)

        prod_precedence = reduce_action.item.production.precedence(cls)
        token_precedence = cls.terminal_precedence(symbol)

        # If both precedences are specified, use priority/associativity
        if prod_precedence is not None and token_precedence is not None:
            prod_assoc, prod_prio = prod_precedence
            token_assoc, token_prio = token_precedence
            if prod_prio > token_prio:
                logger.info('Resolving in favor of reduction because of priority')
                return False
            if prod_prio < token_prio:
                logger.info('Resolving in favor of shift because of priority')
                return True
            if prod_assoc == token_assoc:
                if prod_assoc == 'non':
                    logger.info('Resolving in favor of error because of associativity')
                    raise _ResolveError()
                if prod_assoc == 'left':
                    logger.info('Resolving in favor of reduction because of associativity')
                    return False
                logger.info('Resolving in favor of shift because of associativity')
                return True

        # At least one of those is not specified; use shift
        logger.warning('Could not disambiguate shift/reduce conflict for "%s" on "%s"; using shift', reduce_action.item, symbol)
        cls.nSR += 1
        return True

    @classmethod
    def __resolve_conflicts(cls, logger):
        cls.nSR = 0
        cls.nRR = 0

        for (state, symbol), actions in sorted(cls.__actions__.items()):
            action = actions.pop()
            while actions:
                conflicting = actions.pop()
                try:
                    action = cls.__resolve_conflict(logger, action, conflicting, symbol)
                except _ResolveError:
                    del cls.__actions__[(state, symbol)]
                    break
            else:
                cls.__actions__[(state, symbol)] = action

    @classmethod
    def __resolve_conflict(cls, logger, action1, action2, symbol):
        if isinstance(action2, _Shift):
            action1, action2 = action2, action1

        if isinstance(action1, _Shift):
            # Shift/reduce
            return action1 if cls.__should_prefer_shift(logger, action2, symbol) else action2

        # Reduce/reduce
        logger.warning('Reduce/reduce conflict between "%s" and "%s"', action1.item, action2.item)
        cls.nRR += 1

        # Use the first one to be declared
        for prod in cls.productions():
            if prod == action1.item.production:
                logger.warning('Using "%s', prod)
                return action1
            if prod == action2.item.production:
                logger.warning('Using "%s', prod)
                return action2

        return None

    @classmethod
    def __add_reduce_action(cls, state, symbol, action):
        cls.__actions__.setdefault((state, symbol), []).append(action)

    @classmethod
    def __add_shift_action(cls, state, symbol, action):
        for existing in cls.__actions__.get((state, symbol), []):
            if isinstance(existing, _Shift):
                return
        cls.__actions__.setdefault((state, symbol), []).append(action)

    @classmethod
    def goto(cls, state, symbol):
        return cls._goto[(state, symbol)]

    def _restart_parser(self):
        self.__stack = [_StackItem(self._startState, None, LexerPosition(1, 1))]
        self.__tokens = []
        self.restart_lexer() # pylint: disable=no-member

    @classmethod
    @functools.cache
    def __item_set_closure(cls, items):
        result = set(items)
        while True:
            prev = set(result)
            for item in [item for item in result if not item.should_reduce]:
                symbol = item.production.right[item.dot]
                if symbol not in cls.token_types(): # pylint: disable=no-member
                    terminals = cls.first(*tuple(item.production.right[item.dot + 1:] + [item.terminal]))
                    for prod in (prod for prod in cls.productions() if prod.name == symbol):
                        for terminal in terminals:
                            result.add(_Item(prod, 0, terminal))
            if prev == result:
                break
        return result


class ProductionParser(LRParser, ReLexer): # pylint: disable=R0904
    """
    Parser for productions in other parsers
    """
    def __init__(self, callback, priority, grammar_class, attributes): # pylint: disable=R0915
        self.callback = callback
        self.priority = priority
        self.grammar_class = grammar_class
        self.attributes = attributes

        super().__init__()

    @classmethod
    def prepare(cls, **kwargs): # pylint: disable=R0915
        # Obviously cannot use @production here

        # When mixing async and sync parsers in the same program this may be called twice,
        # because AsyncProductionParser inherits from ProductionParser
        if cls.productions():
            # FIXME remove this, we don't do the async stuff any more
            return

        # DECL -> identifier "->" PRODS
        prod = Production('DECL', cls.DECL)
        prod.add_symbol('LEFT', 'left')
        prod.add_symbol('arrow')
        prod.add_symbol('PRODS', 'prods')
        cls.__productions__.append(prod)

        # LEFT -> identifier
        prod = Production('LEFT', cls.LEFT)
        prod.add_symbol('identifier', 'name')
        cls.__productions__.append(prod)

        # LEFT -> identifier "<" posarg ">"
        prod = Production('LEFT', cls.LEFT)
        prod.add_symbol('identifier', 'name')
        prod.add_symbol('lchev')
        prod.add_symbol('identifier', 'posarg')
        prod.add_symbol('rchev')
        cls.__productions__.append(prod)

        # PRODS -> P
        prod = Production('PRODS', cls.PRODS1)
        prod.add_symbol('P', 'prodlist')
        cls.__productions__.append(prod)

        # PRODS -> PRODS "|" P
        prod = Production('PRODS', cls.PRODS2)
        prod.add_symbol('PRODS', 'prods')
        prod.add_symbol('union')
        prod.add_symbol('P', 'prodlist')
        cls.__productions__.append(prod)

        # P -> P SYM
        prod = Production('P', cls.P1)
        prod.add_symbol('P', 'prodlist')
        prod.add_symbol('SYM', 'sym')
        cls.__productions__.append(prod)

        # P -> ɛ
        prod = Production('P', cls.P2)
        cls.__productions__.append(prod)

        # SYM -> SYMNAME PROPERTIES
        prod = Production('SYM', cls.SYM)
        prod.add_symbol('SYMNAME', 'symname')
        prod.add_symbol('PROPERTIES', 'properties')
        cls.__productions__.append(prod)

        # SYM -> SYMNAME repeat PROPERTIES
        prod = Production('SYM', cls.SYMREP)
        prod.add_symbol('SYMNAME', 'symname')
        prod.add_symbol('repeat', 'repeat')
        prod.add_symbol('PROPERTIES', 'properties')
        cls.__productions__.append(prod)

        # SYM -> SYMNAME repeat lparen identifier rparen PROPERTIES
        prod = Production('SYM', cls.SYMREP)
        prod.add_symbol('SYMNAME', 'symname')
        prod.add_symbol('repeat', 'repeat')
        prod.add_symbol('lparen')
        prod.add_symbol('identifier', 'separator')
        prod.add_symbol('rparen')
        prod.add_symbol('PROPERTIES', 'properties')
        cls.__productions__.append(prod)

        # SYM -> SYMNAME repeat lparen litteral rparen PROPERTIES
        prod = Production('SYM', cls.SYMREP_LIT)
        prod.add_symbol('SYMNAME', 'symname')
        prod.add_symbol('repeat', 'repeat')
        prod.add_symbol('lparen')
        prod.add_symbol('litteral', 'separator')
        prod.add_symbol('rparen')
        prod.add_symbol('PROPERTIES', 'properties')
        cls.__productions__.append(prod)

        # SYMNAME -> identifier
        prod = Production('SYMNAME', cls.SYMNAME1)
        prod.add_symbol('identifier', 'identifier')
        cls.__productions__.append(prod)

        # SYMNAME -> litteral
        prod = Production('SYMNAME', cls.SYMNAME2)
        prod.add_symbol('litteral', 'litteral')
        cls.__productions__.append(prod)

        # PROPERTIES -> ɛ
        prod = Production('PROPERTIES', cls.PROPERTIES1)
        cls.__productions__.append(prod)

        # PROPERTIES -> lchev identifier rchev
        prod = Production('PROPERTIES', cls.PROPERTIES2)
        prod.add_symbol('lchev')
        prod.add_symbol('identifier', 'name')
        prod.add_symbol('rchev')
        cls.__productions__.append(prod)

        super().prepare(**kwargs)

    def on_start_symbol(self, start_symbol):
        (name, posarg), prods = start_symbol
        for prod in prods:
            if prod.name is None:
                prod.name = name
                prod.posarg = posarg
        self.grammar_class.__productions__.extend(prods)

    # Lexer

    @staticmethod
    def ignore(char):
        return char in ' \t\n'

    @token('->')
    def arrow(self, tok):
        pass

    @token('<')
    def lchev(self, tok):
        pass

    @token('>')
    def rchev(self, tok):
        pass

    @token(r'\|')
    def union(self, tok):
        pass

    @token(r'\*|\+|\?')
    def repeat(self, tok):
        pass

    @token(r'\(')
    def lparen(self, tok):
        pass

    @token(r'\)')
    def rparen(self, tok):
        pass

    @token('[a-zA-Z_][a-zA-Z0-9_]*')
    def identifier(self, tok):
        pass

    @token(r'"|\'')
    def litteral(self, tok):
        class StringBuilder: # pylint: disable=missing-class-docstring,too-few-public-methods
            def __init__(self, quotetype):
                self.quotetype = quotetype
                self.chars = []
                self.state = 0
            def feed(self, char):
                if self.state == 0:
                    if char == '\\':
                        self.state = 1
                    elif char == self.quotetype:
                        return 'litteral', ''.join(self.chars)
                    else:
                        self.chars.append(char)
                elif self.state == 1:
                    self.chars.append(char)
                    self.state = 0
                return None
        self.set_consumer(StringBuilder(tok.value))

    # Parser

    def DECL(self, left, prods): # pylint: disable=invalid-name
        name, _ = left
        if name in self.grammar_class.token_types():
            raise GrammarError(f'"{name}" is a token name and cannot be used as non-terminal')
        return (left, prods)

    def LEFT(self, name, posarg=None): # pylint: disable=invalid-name
        return (name, posarg)

    def PRODS1(self, prodlist): # pylint: disable=invalid-name
        return prodlist

    def PRODS2(self, prods, prodlist): # pylint: disable=invalid-name
        prods.extend(prodlist)
        return prods

    def P1(self, sym, prodlist): # pylint: disable=invalid-name
        result = []
        symbol, properties, repeat, sep = sym

        for prod in prodlist:
            if prod.name is None:
                if repeat is None:
                    prod.add_symbol(symbol, name=properties.get('name', None))
                    result.append(prod)
                elif repeat == '?':
                    if sep is not None:
                        raise GrammarError('A separator makes no sense for "?"')
                    self.__add_at_most_one(result, prod, symbol, properties.get('name', None))
                elif repeat in ['*', '+']:
                    self.__add_list(result, prod, symbol, properties.get('name', None), repeat == '*', sep)
            else:
                result.append(prod)

        return result

    def __add_at_most_one(self, productions, prod, symbol, name):
        clone = prod.cloned()
        if name is not None:
            self._wrap_callback_none(name, clone)
        productions.append(clone)

        prod.add_symbol(symbol, name=name)
        productions.append(prod)

    def _wrap_callback_none(self, name, prod):
        previous = prod.callback
        def callback(*args, **kwargs):
            kwargs[name] = None
            return previous(*args, **kwargs)
        prod.callback = callback

    def __add_list(self, productions, prod, symbol, name, allow_empty, sep): # pylint: disable=too-many-arguments,too-many-positional-arguments
        class ListSymbol(metaclass=Singleton): # pylint: disable=missing-class-docstring,too-few-public-methods
            __reprval__ = 'List(%s, "%s")' % (symbol, '*' if allow_empty else '+') # pylint: disable=consider-using-f-string

        if allow_empty:
            clone = prod.cloned()
            self._wrap_callback_empty(name, clone)
            productions.append(clone)

        prod.add_symbol(ListSymbol, name=name)
        productions.append(prod)

        list_prod = Production(ListSymbol, self._wrap_callback_one())
        list_prod.add_symbol(symbol, name='item')
        productions.append(list_prod)

        list_prod = Production(ListSymbol, self._wrap_callback_next())
        list_prod.add_symbol(ListSymbol, name='items')
        if sep is not None:
            list_prod.add_symbol(sep)
        list_prod.add_symbol(symbol, name='item')
        productions.append(list_prod)

    def _wrap_callback_empty(self, name, prod):
        previous = prod.callback
        def callback_empty(*args, **kwargs):
            if name is not None:
                kwargs[name] = []
            return previous(*args, **kwargs)
        prod.callback = callback_empty

    def _wrap_callback_one(self):
        def callback_one(_, item):
            return [item]
        return callback_one

    def _wrap_callback_next(self):
        def callback_next(_, items, item):
            items.append(item)
            return items
        return callback_next

    def P2(self): # pylint: disable=invalid-name
        # 'name' is replaced in on_start_symbol()
        return [Production(None, self.callback, priority=self.priority, attributes=self.attributes)]

    def SYMNAME1(self, identifier): # pylint: disable=invalid-name
        return identifier

    def SYMNAME2(self, litteral): # pylint: disable=invalid-name
        name = litteral
        if name not in self.grammar_class.token_types():
            self.grammar_class.add_token_type(name, lambda s, tok: None, re.escape(name), None)
        return name

    def SYM(self, symname, properties): # pylint: disable=invalid-name
        return (symname, properties, None, None)

    def SYMREP(self, symname, repeat, properties, separator=None): # pylint: disable=invalid-name
        return (symname, properties, repeat, separator)

    def SYMREP_LIT(self, symname, repeat, properties, separator): # pylint: disable=invalid-name
        if separator not in self.grammar_class.token_types():
            self.grammar_class.add_token_type(separator, lambda s, tok: None, re.escape(separator), None)
        return self.SYMREP(symname, repeat, properties, separator)

    def PROPERTIES1(self): # pylint: disable=invalid-name
        return {}

    def PROPERTIES2(self, name): # pylint: disable=invalid-name
        return {'name': name}



# -*- coding: utf-8 -*-

# (c) Jérôme Laheurte 2015-2026
# See LICENSE.txt


class PackageInfo: # pylint: disable=too-few-public-methods
    """
    Information about the package
    """

    version = '2.0.0'
    version_info = map(int, version.split('.'))

    project_name = 'ptk'
    project_url = 'https://codeberg.org/fraca7/ptk'
    download_url = f'https://pypi.python.org/packages/source/p/ptk/ptk-{version}.tar.gz'

    author_name = 'J\u00E9r\u00F4me Laheurte'
    author_email = 'jerome@jeromelaheurte.net'

    short_description = 'LR(1) parsing framework for Python'


version = PackageInfo.version  # pylint: disable=invalid-name
version_info = PackageInfo.version_info  # pylint: disable=invalid-name



#!/usr/bin/env python3

import os, sys, unittest
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

from test_utils import *
from test_lexer import *
from test_grammar import *
from test_parser import *

try:
    import twisted
except ImportError:
    pass
else:
    from test_deferred_lexer import *
    from test_deferred_parser import *


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import codecs

from ptk.meta import PackageInfo

def generateReadme():
    with codecs.getreader('utf-8')(open('README.md.in', 'rb')) as src:
        contents = src.read()
        contents = contents % PackageInfo.__dict__
        with codecs.getwriter('utf-8')(open('README.md', 'wb')) as dst:
            dst.write(contents)


def prepare():
    generateReadme()


if __name__ == '__main__':
    prepare()



#!/usr/bin/env python3

import base, unittest, copy
from ptk.utils import Singleton, chars


class SingletonUnderTest(metaclass=Singleton):
    __reprval__ = '$'

    def __init__(self):
        self.value = 42


class SingletonTestCase(unittest.TestCase):
    def test_instance(self):
        self.assertFalse(isinstance(SingletonUnderTest, type))

    def test_init(self):
        self.assertEqual(SingletonUnderTest.value, 42)

    def test_order(self):
        values = ['spam', SingletonUnderTest]
        values.sort()
        self.assertEqual(values, [SingletonUnderTest, 'spam'])

    def test_copy(self):
        self.assertTrue(copy.copy(SingletonUnderTest) is SingletonUnderTest)

    def test_deepcopy(self):
        self.assertTrue(copy.deepcopy(SingletonUnderTest) is SingletonUnderTest)

    def test_repr(self):
        self.assertEqual(repr(SingletonUnderTest), '$')

    def test_eq(self):
        self.assertNotEqual(SingletonUnderTest, SingletonUnderTest.__class__())


class CharsTestCase(unittest.TestCase):
    def test_str(self):
        self.assertTrue('*' in chars('*'))

    def test_bytes(self):
        for byte in b'*':
            self.assertTrue(byte in chars('*'))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

import base, unittest

from ptk.grammar import Production, GrammarError, Grammar, production
from ptk.lexer import token
from ptk.parser import ProductionParser
from ptk.utils import callback_by_name


class GrammarUnderTest(Grammar):
    tokens = set()

    @classmethod
    def add_token_type(cls, name, callback, regex, types=None):
        if types is None:
            cls.tokens.add(name)
        else:
            for name in types:
                cls.tokens.add(name)

    @classmethod
    def token_types(cls):
        return cls.tokens

    @classmethod
    def _create_production_parser(cls, name, priority, attrs):
        return ProductionParser(callback_by_name(name), priority, cls, attrs)


class ProductionParserTestCase(unittest.TestCase):
    def setUp(self):
        class TestGrammar(GrammarUnderTest):
            tokens = set()
        self.parser = ProductionParser(None, None, TestGrammar, dict())
        self.grammar_class = TestGrammar

    def assertHasProduction(self, prods, prod):
        for aProduction in prods:
            if prod == (aProduction.name, aProduction.right):
                return
        self.fail('Production %s not found in %s' % (prod, prods))

    def _parse(self, string):
        self.parser.parse(string)
        return self.grammar_class.productions()

    def test_production_name(self):
        prod, = self._parse('test -> A')
        self.assertEqual(prod.name, 'test')

    def test_production_callback(self):
        prod, = self._parse('test -> A')
        self.assertEqual(prod.callback, None)

    def test_empty_production(self):
        prod, = self._parse('test -> ')
        self.assertEqual(prod.right, [])

    def test_empty_end(self):
        prod1, prod2 = self._parse('test -> A | ')
        self.assertEqual(prod1.right, ['A'])
        self.assertEqual(prod2.right, [])

    def test_empty_start(self):
        prod1, prod2 = self._parse('test -> | A')
        self.assertEqual(prod1.right, [])
        self.assertEqual(prod2.right, ['A'])

    def test_order(self):
        prod, = self._parse('test -> A B')
        self.assertEqual(prod.right, ['A', 'B'])

    def test_escape_litteral(self):
        prod, = self._parse(r'test -> "spam\"foo"')
        self.assertEqual(prod.right, [r'spam"foo'])

    def _findListSym(self, prods):
        for prod in prods:
            if prod.name != 'test':
                return prod.name
        self.fail('Cannot find list symbol in %s' % repr(prods))

    def test_list(self):
        prods = self._parse('test -> A* "+"')
        self.assertEqual(len(prods), 4, repr(prods))
        listSym = self._findListSym(prods)

        self.assertHasProduction(prods, ('test', ['+']))
        self.assertHasProduction(prods, (listSym, ['A']))
        self.assertHasProduction(prods, (listSym, [listSym, 'A']))
        self.assertHasProduction(prods, ('test', [listSym, '+']))

    def test_list_not_empty(self):
        prods = self._parse('test -> A+')
        listSym = self._findListSym(prods)
        self.assertEqual(len(prods), 3, repr(prods))

        self.assertHasProduction(prods, (listSym, ['A']))
        self.assertHasProduction(prods, (listSym, [listSym, 'A']))
        self.assertHasProduction(prods, ('test', [listSym]))

    def test_list_with_separator(self):
        prods = self._parse('test -> A+(pipe)')
        self.assertEqual(len(prods), 3, repr(prods))
        listSym = self._findListSym(prods)

        self.assertHasProduction(prods, (listSym, ['A']))
        self.assertHasProduction(prods, (listSym, [listSym, 'pipe', 'A']))
        self.assertHasProduction(prods, ('test', [listSym]))

    def test_list_with_litteral_separator(self):
        prods = self._parse('test -> A+("|")')
        self.assertEqual(len(prods), 3, repr(prods))
        listSym = self._findListSym(prods)

        self.assertHasProduction(prods, (listSym, ['A']))
        self.assertHasProduction(prods, (listSym, [listSym, '|', 'A']))
        self.assertHasProduction(prods, ('test', [listSym]))

    def test_atmostone(self):
        prods = self._parse('test -> A?')
        self.assertEqual(len(prods), 2)
        self.assertHasProduction(prods, ('test', []))
        self.assertHasProduction(prods, ('test', ['A']))


class ProductionTestCase(unittest.TestCase):
    def setUp(self):
        # A -> B<b> C
        self.production = Production('A', self.callback)
        self.production.add_symbol('B', 'b')
        self.production.add_symbol('C')
        self.calls = list()

    def callback(self, grammar, **kwargs):
        self.calls.append(kwargs)
        return 42

    def test_duplicate(self):
        try:
            self.production.add_symbol('D', 'b')
        except GrammarError:
            pass
        else:
            self.fail()

    def test_kwargs(self):
        cb, kwargs = self.production.apply([1, 2], None)
        cb(self, **kwargs)
        self.assertEqual(self.calls, [{'b': 1}])


class GrammarTestCase(unittest.TestCase):
    def test_production(self):
        class G(GrammarUnderTest):
            @production('A -> B')
            def a(self):
                pass
        prod, = G.productions()
        self.assertEqual(prod.name, 'A')
        self.assertEqual(prod.right, ['B'])

    def test_start_symbol(self):
        class G(GrammarUnderTest):
            @production('A -> B')
            def a(self):
                pass
            @production('C -> D')
            def c(self):
                pass

        grammar = G()
        self.assertEqual(grammar.startSymbol, 'A')

    def test_duplicate_name(self):
        class G(GrammarUnderTest):
            @production('A -> B')
            def a1(self):
                pass
            @production('A -> C')
            def a2(self):
                pass
        prod1, prod2 = G.productions()
        self.assertEqual(prod1.name, 'A')
        self.assertEqual(prod1.right, ['B'])
        self.assertEqual(prod2.name, 'A')
        self.assertEqual(prod2.right, ['C'])

    def test_token_name(self):
        try:
            class G(GrammarUnderTest):
                tokens = set(['spam'])
                @production('spam -> spam')
                def spam(self):
                    pass
        except GrammarError:
            pass
        else:
            self.fail()

    def test_duplicate_production_1(self):
        class G(GrammarUnderTest):
            @production('A -> B|B')
            def a(self):
                pass

        try:
            G()
        except GrammarError:
            pass
        else:
            self.fail()

    def test_duplicate_production_2(self):
        class G(GrammarUnderTest):
            @production('A -> B')
            def a1(self):
                pass
            @production('A -> B')
            def a2(self):
                pass

        try:
            G()
        except GrammarError:
            pass
        else:
            self.fail()


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

import base, unittest

from ptk.parser import LRParser, ParseError, left_assoc, right_assoc, non_assoc
from ptk.lexer import ReLexer, token, EOF
from ptk.grammar import production


class TestedParser(LRParser, ReLexer):
    def __init__(self, *args, **kwargs):
        self.seen = list()
        super().__init__(*args, **kwargs)

    @token('[1-9][0-9]*')
    def number(self, tok):
        tok.value = int(tok.value)

    @production('E -> E<left> "+" E<right>')
    def sum(self, left, right):
        self.seen.append(('+', left, right))
        return left + right

    @production('E -> E<left> "*" E<right>')
    def mult(self, left, right):
        self.seen.append(('*', left, right))
        return left * right

    @production('E -> number<n>')
    def litteral(self, n):
        return n

    def on_start_symbol(self, sentence):
        pass


class DefaultShiftReduceTestCase(unittest.TestCase):
    def setUp(self):
        self.parser = TestedParser()

    def test_shift(self):
        self.parser.parse('2+3*4')
        self.assertEqual(self.parser.seen, [('*', 3, 4), ('+', 2, 12)])


class DefaultReduceReduceTestCase(unittest.TestCase):
    def setUp(self):
        class Parser(LRParser, ReLexer):
            def __init__(self, *args, **kwargs):
                self.seen = list()
                super().__init__(*args, **kwargs)
            @token('[a-zA-Z]+')
            def word(self, tok):
                pass
            @production('sequence -> maybeword | sequence word |')
            def seq(self):
                self.seen.append('seq')
            @production('maybeword -> word |')
            def maybe(self):
                self.seen.append('maybe')
            def on_start_symbol(self, sentence):
                pass

        self.parser = Parser()

    def test_reduce(self):
        self.parser.parse('')
        self.assertEqual(self.parser.seen, ['seq'])


class LeftAssociativityTestCase(unittest.TestCase):
    def setUp(self):
        @left_assoc('+')
        class Parser(TestedParser):
            pass

        self.parser = Parser()

    def test_assoc(self):
        self.parser.parse('1+2+3')
        self.assertEqual(self.parser.seen, [('+', 1, 2), ('+', 3, 3)])


class RightAssociativityTestCase(unittest.TestCase):
    def setUp(self):
        @right_assoc('+')
        class Parser(TestedParser):
            pass

        self.parser = Parser()

    def test_assoc(self):
        self.parser.parse('1+2+3')
        self.assertEqual(self.parser.seen, [('+', 2, 3), ('+', 1, 5)])


class PrecedenceTestCase(unittest.TestCase):
    def setUp(self):
        @left_assoc('+')
        @left_assoc('*')
        @non_assoc('<')
        class Parser(TestedParser):
            @production('E -> E "<" E')
            def inf(self):
                pass

        self.parser = Parser()

    def test_shift(self):
        self.parser.parse('2+3*4')
        self.assertEqual(self.parser.seen, [('*', 3, 4), ('+', 2, 12)])

    def test_reduce(self):
        self.parser.parse('2*3+4')
        self.assertEqual(self.parser.seen, [('*', 2, 3), ('+', 6, 4)])

    def test_error(self):
        try:
            self.parser.parse('2 < 3 < 4')
        except ParseError:
            pass
        else:
            self.fail()


class OverridePrecedenceTestCase(unittest.TestCase):
    def setUp(self):
        @left_assoc('+')
        @left_assoc('mult')
        class Parser(LRParser, ReLexer):
            def __init__(self, *args, **kwargs):
                self.seen = list()
                super().__init__(*args, **kwargs)
            @token('[1-9][0-9]*')
            def number(self, tok):
                tok.value = int(tok.value)
            @production('E -> E<left> "+" E<right>')
            def sum(self, left, right):
                self.seen.append(('+', left, right))
                return left + right
            @production('E -> E<left> "*" E<right>', priority='mult')
            def mult(self, left, right):
                self.seen.append(('*', left, right))
                return left * right
            @production('E -> number<n>')
            def litteral(self, n):
                return n
            def on_start_symbol(self, sentence):
                pass

        self.parser = Parser()

    def test_shift(self):
        self.parser.parse('2+3*4')
        self.assertEqual(self.parser.seen, [('*', 3, 4), ('+', 2, 12)])

    def test_reduce(self):
        self.parser.parse('2*3+4')
        self.assertEqual(self.parser.seen, [('*', 2, 3), ('+', 6, 4)])


class LRTestCase(unittest.TestCase):
    def setUp(self):
        class Parser(LRParser, ReLexer):
            @token('a')
            def identifier(self, tok):
                pass
            @production('S -> G "=" D | D')
            def s(self):
                pass
            @production('G -> "*" D | identifier')
            def g(self):
                pass
            @production('D -> G')
            def d(self):
                pass

        self.parser = Parser()

    def test_no_conflicts(self):
        self.assertEqual(self.parser.nSR, 0)


class ListTestCase(unittest.TestCase):
    def setUp(self):
        class Parser(LRParser, ReLexer):
            def __init__(self, testCase):
                super().__init__()
                self.testCase = testCase
            @token('[a-z]+')
            def identifier(self, tok):
                pass
            @production('L -> identifier*<tokens>')
            def idlist(self, tokens):
                return tokens
            def on_start_symbol(self, symbol):
                self.testCase.seen = symbol

        self.parser = Parser(self)
        self.seen = None

    def test_empty(self):
        self.parser.parse('')
        self.assertEqual(self.seen, [])

    def test_items(self):
        self.parser.parse('a b c')
        self.assertEqual(self.seen, ['a', 'b', 'c'])


class NonEmptyListTestCase(unittest.TestCase):
    def setUp(self):
        class Parser(LRParser, ReLexer):
            def __init__(self, testCase):
                super().__init__()
                self.testCase = testCase
            @token('[a-z]+')
            def identifier(self, tok):
                pass
            @production('L -> identifier+<tokens>')
            def idlist(self, tokens):
                return tokens
            def newSentence(self, symbol):
                self.testCase.seen = symbol

        self.parser = Parser(self)
        self.seen = None

    def test_empty(self):
        try:
            self.parser.parse('')
        except ParseError:
            pass
        else:
            self.fail('Got %s' % self.seen)

    def test_items(self):
        self.parser.parse('a b c')
        self.assertEqual(self.seen, ['a', 'b', 'c'])


class SeparatorListTestCase(unittest.TestCase):
    def setUp(self):
        class Parser(LRParser, ReLexer):
            def __init__(self, testCase):
                super().__init__()
                self.testCase = testCase
            @token('[a-z]+')
            def identifier(self, tok):
                pass
            @production('L -> identifier+("|")<tokens>')
            def idlist(self, tokens):
                return tokens
            def on_start_symbol(self, symbol):
                self.testCase.seen = symbol

        self.parser = Parser(self)
        self.seen = None

    def test_items(self):
        self.parser.parse('a | b | c')
        self.assertEqual(self.seen, ['a', 'b', 'c'])


class AtMostOneTestCase(unittest.TestCase):
    def setUp(self):
        class Parser(LRParser, ReLexer):
            def __init__(self, testCase):
                super().__init__()
                self.testCase = testCase
            @token('[a-z]+')
            def identifier(self, tok):
                pass
            @production('L -> identifier?<tok>')
            def idlist(self, tok):
                return tok
            def on_start_symbol(self, symbol):
                self.testCase.seen = symbol

        self.parser = Parser(self)
        self.seen = 1

    def test_none(self):
        self.parser.parse('')
        self.assertEqual(self.seen, None)

    def test_value(self):
        self.parser.parse('a')
        self.assertEqual(self.seen, 'a')

    def test_error(self):
        try:
            self.parser.parse('a b')
        except ParseError:
            pass
        else:
            self.fail('Got %s' % self.seen)


class InheritanceTestCase(unittest.TestCase):
    def setUp(self):
        class Parser(LRParser, ReLexer):
            def __init__(self):
                self.seen = None
                super().__init__()
            @token('[0-9]')
            def digit(self, tok):
                tok.value = int(tok.value)
            @production('E -> digit<d>')
            def start(self, d):
                pass
            def on_start_symbol(self, symbol):
                self.seen = symbol

        class Child(Parser):
            def start(self, d):
                return d

        self.parser = Child()

    def test_override(self):
        self.parser.parse('4')
        self.assertEqual(self.parser.seen, 4)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

# (c) Jérôme Laheurte 2015-2026
# See LICENSE.txt

"""
Miscellaneous utilities.
"""

import functools


class Singleton(type):
    """
    Singleton metaclass
    """
    def __new__(mcs, name, bases, attrs):
        cls = type.__new__(mcs, name, bases, attrs)
        cls.__eq__ = lambda self, other: other is self
        cls.__lt__ = lambda self, other: not other is self
        cls.__copy__ = lambda self: self
        cls.__deepcopy__ = lambda self, memo: self
        cls.__repr__ = lambda self: self.__reprval__
        cls.__len__ = lambda self: len(self.__reprval__)
        cls.__hash__ = lambda self: hash(id(self))
        return functools.total_ordering(cls)()


def callback_by_name(func_name): # pylint: disable=missing-function-docstring
    def _wrapper(instance, *args, **kwargs):
        return getattr(instance, func_name)(*args, **kwargs)
    return _wrapper


def chars(s): # pylint: disable=missing-function-docstring
    return (s, s.encode('ascii')[0])



# -*- coding: utf-8 -*-

# (c) Jérôme Laheurte 2015-2026
# See LICENSE.txt

"""
Context-free grammars objects. To define a grammar, inherit the
Grammar class and define a method decorated with 'production' for each
production.
"""

import copy
import functools
import inspect
import logging
import warnings

from ptk.lexer import EOF, _LexerMeta
from ptk.utils import Singleton


class Epsilon(metaclass=Singleton): # pylint: disable=too-few-public-methods
    """
    Empty production
    """
    __reprval__ = '\u03B5'


class GrammarError(Exception):
    """
    Generic grammar error, like duplicate production.
    """


class GrammarParseError(GrammarError):
    """
    Syntax error in a production specification.
    """


@functools.total_ordering
class Production:
    """
    Production object
    """
    def __init__(self, name, callback, priority=None, attributes=None):
        self.name = name
        self.posarg = None
        self.callback = callback
        self.right = []
        self.attributes = attributes or {}
        self.__priority = priority
        # position => id
        self.__ids = {} # pylint: disable=unused-private-member

    def add_symbol(self, identifier, name=None):
        """
        Append a symbol to the production's right side.
        """
        if name is not None:
            if name in self.__ids.values():
                raise GrammarParseError(f'Duplicate identifier name "{name}"')
            self.__ids[len(self.right)] = name
        self.right.append(identifier)

    def addSymbol(self, identifier, name=None): # pylint: disable=invalid-name
        warnings.warn('addSymbol is deprecated in favor of add_symbol', DeprecationWarning)
        self.add_symbol(identifier, name=name)

    def cloned(self):
        prod = Production(self.name, self.callback, self.__priority)
        prod.right = list(self.right)
        prod.__ids = dict(self.__ids) # pylint: disable=protected-access,unused-private-member
        return prod

    def apply(self, args, position):
        kwargs = {name: args[index] for index, name in self.__ids.items()}
        if self.posarg is not None:
            kwargs[self.posarg] = position
        return self.callback, kwargs

    def rightmost_terminal(self, grammar):
        """
        Returns the rightmost terminal, or None if there is none
        """
        for symbol in reversed(self.right):
            if symbol in grammar.token_types():
                return symbol
        return None

    def rightmostTerminal(self, grammar): # pylint: disable=invalid-name
        warnings.warn('rightmostTerminal is deprecated in favor of rightmost_terminal', DeprecationWarning)
        return self.rightmost_terminal(grammar)

    def precedence(self, grammar):
        """
        Returns the production's priority (specified through the
        'priority' keyword argument to the 'production' decorator), or
        if there is none, the priority for the rightmost terminal.
        """
        if self.__priority is not None:
            return grammar.terminal_precedence(self.__priority)
        symbol = self.rightmost_terminal(grammar)
        if symbol is not None:
            return grammar.terminal_precedence(symbol)
        return None

    def __eq__(self, other):
        return (self.name, self.right) == (other.name, other.right)

    def __lt__(self, other):
        return (self.name, self.right) < (other.name, other.right)

    def __repr__(self): # pragma: no cover
        return '%s -> %s' % (self.name, ' '.join([repr(p) for p in self.right]) if self.right else repr(Epsilon)) # pylint: disable=consider-using-f-string

    def __hash__(self):
        return hash((self.name, tuple(self.right)))


# Same remark as in lexer.py.
_PRODREGISTER = []


class _GrammarMeta(_LexerMeta):
    def __new__(mcs, name, bases, attrs):
        global _PRODREGISTER # pylint: disable=W0603
        try:
            attrs['__productions__'] = []
            attrs['__precedence__'] = []
            attrs['__prepared__'] = False
            attrs['__lrstates__'] = []
            klass = super().__new__(mcs, name, bases, attrs)
            for func, string, priority, attributes in _PRODREGISTER:
                parser = klass._create_production_parser(func.__name__, priority, attributes) # pylint: disable=W0212
                parser.parse(string)
            return klass
        finally:
            _PRODREGISTER = []


def production(prod, priority=None, **kwargs):
    def _wrap(func):
        if any(func.__name__ == aFunc.__name__ and func != aFunc for aFunc, _, _, _ in _PRODREGISTER):
            raise TypeError(f'Duplicate production method name "{func.__name__}"')
        _PRODREGISTER.append((func, prod, priority, kwargs))
        return func
    return _wrap


class Grammar(metaclass=_GrammarMeta):
    """
    Base class for a context-free grammar
    """

    __productions__ = [] # Make pylint happy
    __precedence__ = []
    __prepared__ = False

    startSymbol = None

    def __init__(self):
        # pylint: disable=R0912
        super().__init__()
        if not self.__prepared__:
            self.prepare()

    @classmethod
    def prepare(cls):
        cls.startSymbol = cls._default_start_symbol() if cls.startSymbol is None else cls.startSymbol

        productions = set()
        for prod in cls.productions():
            if prod in productions:
                raise GrammarError(f'Duplicate production "{prod}"')
            productions.add(prod)

        cls.__all_firsts__ = cls.__compute_firsts()

        logger = logging.getLogger('Grammar')
        productions = cls.productions()
        max_width = max(len(prod.name) for prod in productions)
        for prod in productions:
            logger.debug('%%- %ds -> %%s' % max_width, prod.name, ' '.join([repr(name) for name in prod.right]) if prod.right else Epsilon) # pylint: disable=consider-using-f-string,logging-not-lazy

        cls.__prepared__ = True

    @classmethod
    def __compute_firsts(cls):
        # pylint: disable=too-many-nested-blocks
        all_firsts = {symbol: set([symbol]) for symbol in cls.token_types()|set([EOF])} # pylint: disable=no-member
        while True:
            prev = copy.deepcopy(all_firsts)
            for nonterminal in cls.nonterminals():
                for prod in cls.productions():
                    if prod.name == nonterminal:
                        if prod.right:
                            for symbol in prod.right:
                                first = all_firsts.get(symbol, set())
                                all_firsts.setdefault(nonterminal, set()).update(first)
                                if Epsilon not in first:
                                    break
                            else:
                                all_firsts.setdefault(nonterminal, set()).add(Epsilon)
                        else:
                            all_firsts.setdefault(nonterminal, set()).add(Epsilon)
            if prev == all_firsts:
                break
        return all_firsts

    @classmethod
    def _default_start_symbol(cls):
        return cls.productions()[0].name

    @classmethod
    def productions(cls):
        """
        Returns all productions
        """
        productions = []
        for base in inspect.getmro(cls):
            if issubclass(base, Grammar):
                productions.extend(base.__productions__)
        return productions

    @classmethod
    def nonterminals(cls):
        """
        Return all non-terminal symbols
        """
        result = set()
        for prod in cls.productions():
            result.add(prod.name)
            for symbol in prod.right:
                if symbol not in cls.token_types(): # pylint: disable=no-member
                    result.add(symbol)
        return result

    @classmethod
    def precedences(cls):
        precedences = []
        for base in inspect.getmro(cls):
            if issubclass(base, Grammar):
                precedences.extend(base.__precedence__)
        return precedences

    @classmethod
    def terminal_precedence(cls, symbol):
        for index, (associativity, terminals) in enumerate(cls.precedences()):
            if symbol in terminals:
                return associativity, index
        return None

    @classmethod
    def terminalPrecedence(cls, symbol): # pylint: disable=invalid-name
        warnings.warn('terminalPrecedence is deprecated in favor of terminal_precedence', DeprecationWarning)
        return cls.terminal_precedence(symbol)

    @classmethod
    @functools.cache
    def first(cls, *symbols):
        """
        Returns the first set for a group of symbols
        """
        first = set()
        for symbol in symbols:
            rfirst = cls.__all_firsts__[symbol]
            first |= {a for a in rfirst if a is not Epsilon}
            if Epsilon not in rfirst:
                break
        else:
            first.add(Epsilon)
        return first



# -*- coding: utf-8 -*-

import sys
import os
import shlex

sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', '..'))
from ptk.meta import version, PackageInfo

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
]

templates_path = ['_templates']
source_suffix = '.rst'
master_doc = 'index'

project = PackageInfo.project_name
copyright = '2019, %s' % PackageInfo.author_name
author = PackageInfo.author_name
release = version
language = None
exclude_patterns = []
pygments_style = 'sphinx'
todo_include_todos = False

html_theme = 'alabaster'
html_static_path = ['_static']
htmlhelp_basename = '{name}doc'.format(name=author)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Simple four operations calculator.
"""

import operator

from ptk.lexer import ReLexer, token
from ptk.parser import LRParser, left_assoc, production, ParseError


@left_assoc('+', '-')
@left_assoc('*', '/')
class SimpleCalc(LRParser, ReLexer):
    def on_start_symbol(self, result):
        print('== Result:', result)

    # Lexer
    def ignore(self, char):
        return char in [' ', '\t']

    @token(r'[1-9][0-9]*')
    def number(self, tok):
        tok.value = int(tok.value)

    # Parser

    @production('E -> "-" E<value>', priority='*')
    def minus(self, value):
        print('== Neg: - %d' % value)
        return -value

    @production('E -> "(" E<value> ")"')
    def paren(self, value):
        return value

    @production('E -> number<number>')
    def litteral(self, number):
        return number

    @production('E -> E<left> "+"<op> E<right>')
    @production('E -> E<left> "-"<op> E<right>')
    @production('E -> E<left> "*"<op> E<right>')
    @production('E -> E<left> "/"<op> E<right>')
    def binaryop(self, left, op, right):
        print('Binary operation: %s %s %s' % (left, op, right))
        return {
            '+': operator.add,
            '-': operator.sub,
            '*': operator.mul,
            '/': operator.floordiv
            }[op](left, right)


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=logging.WARNING, format='%(asctime)-15s %(levelname)-8s %(name)-15s %(message)s')

    print('Enter an arithmetic expression.')

    parser = SimpleCalc()
    while True:
        try:
            line = input('> ')
        except (KeyboardInterrupt, EOFError):
            print()
            break
        try:
            parser.parse(line)
        except ParseError as exc:
            print('Parse error: %s' % exc)
            print('Expected %s' % exc.expecting())



# -*- coding: utf-8 -*-

# (c) Jérôme Laheurte 2015-2026
# See LICENSE.txt

import inspect
import re
import collections
import warnings

from ptk.utils import Singleton, callback_by_name, chars


LexerPosition = collections.namedtuple('_LexerPosition', ['column', 'line'])

# In Python 3 we'd use __prepare__ and an ordered dict...
_TOKREGISTER = []


class _LexerMeta(type):
    def __new__(mcs, name, bases, attrs):
        global _TOKREGISTER # pylint: disable=W0603
        try:
            attrs['__tokens__'] = (set(), []) # Set of token names, list of (rx, callback, defaultType)
            klass = super().__new__(mcs, name, bases, attrs)
            for func, rx, toktypes in _TOKREGISTER:
                klass.add_token_type(func.__name__, callback_by_name(func.__name__), rx, toktypes)
            return klass
        finally:
            _TOKREGISTER = []


def token(rx, types=None):
    """
    The method decorator for tokens
    """
    def _wrap(func):
        if any(func.__name__ == aFunc.__name__ and func != aFunc for aFunc, _, _ in _TOKREGISTER):
            raise TypeError(f'Duplicate token method name "{func.__name__}"')
        _TOKREGISTER.append((func, rx, types))
        return func
    return _wrap


class SkipToken(Exception):
    """
    Raise this from your consumer to ignore the token.
    """


class LexerError(Exception):
    """
    Unrecognized token in input

    :ivar lineno: Line in input
    :ivar colno: Column in input
    """
    def __init__(self, char, pos):
        super().__init__(f'Unrecognized token {repr(char)}')
        self.position = pos

    # Getters for compatibility with <1.3.8
    @property
    def colno(self):
        """
        Column
        """
        return self.position.column
    @property
    def lineno(self):
        """
        Line
        """
        return self.position.line


class EOF(metaclass=Singleton):
    """
    End symbol
    """

    __reprval__ = '$'

    @property
    def type(self):
        """Read-only attribute for Token duck-typing"""
        return self
    @property
    def value(self):
        """Read-only attribute for Token duck-typing"""
        return self


class LexerBase(metaclass=_LexerMeta):
    """
    This defines the interface for lexer classes. For concrete
    implementations, see :py:class:`ProgressiveLexer` and
    :py:class:`ReLexer`.
    """

    Token = collections.namedtuple('Token', ['type', 'value', 'position'])

    # Shut up pychecker. Those are actually set by the metaclass.
    __tokens__ = ()

    class _MutableToken: # pylint: disable=too-few-public-methods
        def __init__(self, type_, value, position):
            self.type = type_
            self.value = value
            self.position = position

        def token(self):
            """Returns the unmutable equivalent"""
            return EOF if EOF in [self.type, self.value] else LexerBase.Token(self.type, self.value, self.position)

    def __init__(self):
        super().__init__()

        self._pos = None
        self._consumer = None

        self.restart_lexer()

    def restart_lexer(self, reset_pos=True):
        if reset_pos:
            self._pos = LexerPosition(column=1, line=1)
            self._input = []
        self._consumer = None

    def restartLexer(self, resetPos=True): # pylint: disable=invalid-name
        warnings.warn('restartLexer is deprecated in favor of restart_lexer', DeprecationWarning)
        self.restart_lexer(reset_pos=resetPos)

    def position(self):
        """
        :return: The current position in stream as a 2-tuple (column, line).
        """
        return self._pos

    def advance_column(self, count=1):
        """
        Advances the current position by *count* columns.
        """
        self._pos = self._pos._replace(column=self._pos.column + count)

    def advanceColumn(self, count=1): # pylint: disable=invalid-name
        warnings.warn('advanceColumn is deprecated in favor of advance_column', DeprecationWarning)
        self.advance_column(count=count)

    def advance_line(self, count=1):
        """
        Advances the current position by *count* lines.
        """
        self._pos = self._pos._replace(column=1, line=self._pos.line + count)

    def advanceLine(self, count=1): # pylint: disable=invalid-name
        warnings.warn('advanceLine is deprecated in favor of advance_line', DeprecationWarning)
        self.advance_line(count=count)

    @staticmethod
    def ignore(char):
        """
        Override this to ignore characters in input stream. The
        default is to ignore spaces and tabs.

        :param char: The character to test
        :return: True if *char* should be ignored
        """
        return char in chars(' ') + chars('\t')

    def set_consumer(self, consumer):
        """
        Sets the current consumer. A consumer is an object with a
        *feed* method; all characters seen on the input stream after
        the consumer is set are passed directly to it. When the *feed*
        method returns a 2-tuple (type, value), the corresponding
        token is generated and the consumer reset to None. This may be
        handy to parse tokens that are not easily recognized by a
        regular expression but easily by code; for instance the
        following lexer recognizes C strings without having to use
        negative lookahead:

        .. code-block:: python

           class MyLexer(ReLexer):
               @token('"')
               def cstring(self, tok):
                   class CString:
                       def __init__(self):
                           self.state = 0
                           self.value = StringIO.StringIO()
                       def feed(self, char):
                           if self.state == 0:
                               if char == '"':
                                   return 'cstring', self.value.getvalue()
                               if char == '\\\\':
                                   self.state = 1
                               else:
                                   self.value.write(char)
                           elif self.state == 1:
                               self.value.write(char)
                               self.state = 0
                   self.set_consumer(CString())

        You can also raise SkipToken instead of returning a token if it
        is to be ignored (comments).
        """
        self._consumer = consumer

    def setConsumer(self, consumer): # pylint: disable=invalid-name
        warnings.warn('setConsumer is deprecated in favor of set_consumer', DeprecationWarning)
        self.set_consumer(consumer)

    def consumer(self):
        return self._consumer

    def parse(self, string): # pragma: no cover
        """
        Parses the whole *string*; returns the start symbol semantic value
        """
        raise NotImplementedError

    def new_token(self, tok): # pragma: no cover
        """
        This method will be invoked as soon as a token is recognized on input.

        :param tok: The token. This is a named tuple with *type* and *value* attributes.
        """
        raise NotImplementedError

    def newToken(self, tok): # pylint: disable=invalid-name
        warnings.warn('newToken is deprecated in favor of new_token', DeprecationWarning)
        self.new_token(tok)

    @classmethod
    def add_token_type(cls, name, callback, regex, types=None):
        for type_name in [name] if types is None else types:
            if type_name is not EOF:
                cls.__tokens__[0].add(type_name)
        cls.__tokens__[1].append((regex, callback, name if types is None else None))

    @classmethod
    def _all_tokens(cls):
        tokens = (set(), [])
        for base in inspect.getmro(cls):
            if issubclass(base, LexerBase):
                tokens[0].update(base.__tokens__[0])
                tokens[1].extend(base.__tokens__[1])
        return tokens

    @classmethod
    def token_types(cls):
        """
        :return: the set of all token names, as strings.
        """
        return cls._all_tokens()[0]

    @classmethod
    def tokenTypes(cls): # pylint: disable=invalid-name
        warnings.warn('tokenTypes is deprecated in favor of token_types', DeprecationWarning)
        return cls.token_types()


class ReLexer(LexerBase): # pylint: disable=W0223
    """
    Concrete lexer based on Python regular expressions.
    """
    def __init__(self):
        self._regexes = []
        for rx, callback, default_type in self._all_tokens()[1]:
            crx = re.compile((b'^' if isinstance(rx, bytes) else '^') + rx)
            self._regexes.append((crx, callback, default_type))
        super().__init__()

    def _parse(self, string, pos):
        # pylint: disable=too-many-nested-blocks
        while pos < len(string):
            char = string[pos]
            try:
                if self.consumer() is None:
                    if self.ignore(char):
                        pos += 1
                        continue
                    pos = self._find_match(string, pos)
                else:
                    try:
                        tok = self.consumer().feed(char)
                    except SkipToken:
                        self.set_consumer(None)
                    else:
                        if tok is not None:
                            self.set_consumer(None)
                            if tok[0] is not None:
                                self.new_token(self.Token(*tok, self.position()))
                    pos += 1
            finally:
                if char in chars('\n'):
                    self.advance_line()
                else:
                    self.advance_column()
        return pos

    def parse(self, string):
        try:
            self._parse(string, 0)
            return self.new_token(EOF)
        except LexerError:
            self.restart_lexer()
            raise

    def _find_match(self, string, pos):
        match = None
        matchlen = 0
        pos2d = self.position()

        for rx, callback, default_type in self._regexes:
            mtc = rx.match(string[pos:])
            if mtc:
                value = mtc.group(0)
                if len(value) > matchlen:
                    match = value, callback, default_type
                    matchlen = len(value)

        if match:
            value, callback, default_type = match
            tok = self._MutableToken(default_type, value, pos2d)
            callback(self, tok)
            pos += matchlen
            if self.consumer() is None and tok.type is not None:
                self.new_token(tok.token())
            self.advance_column(matchlen - 1)
            return pos

        raise LexerError(self._guess_token(string, pos), pos2d)

    def _guess_token(self, string, pos):
        start = pos
        while True:
            pos += 1
            if pos == len(string) or self.ignore(string[pos]):
                break

            for rx, _, _ in self._regexes:
                mtc = rx.match(string[pos:])
                if mtc:
                    break
            else:
                continue
            break
        return string[start:pos]



#!/usr/bin/env python3

import io
import base, unittest

from ptk.lexer import LexerError, token, EOF, LexerPosition
from ptk.lexer import ReLexer


class LexerUnderTestMixin(object):
    def __init__(self, testCase):
        self.testCase = testCase
        super().__init__()

    def new_token(self, tok):
        self.testCase.feed(tok)


class LexerTestCase(unittest.TestCase):
    def setUp(self):
        self.tokens = list()

    def feed(self, tok):
        if tok is EOF:
            self.tokens = tuple(self.tokens) # So that any more tokens will raise an exception
        else:
            self.tokens.append(tok)

    def doLex(self, inputString):
        self.lexer.parse(inputString)
        return self.tokens


class ReLexerTestCase(LexerTestCase):
    lexerClass = ReLexer


class LexerBasicTestCaseMixin(object):
    def setUp(self):
        super().setUp()
        class TestedLexer(LexerUnderTestMixin, self.lexerClass):
            @token('[a-zA-Z]+')
            def ID(self, tok):
                pass
            @token('[0-9]+')
            def NUMBER(self, tok):
                tok.value = int(tok.value)
            @token('\n', types=[EOF])
            def EOL(self, tok):
                tok.type = EOF
            @token('0x[a-fA-F0-9]+', types=['NUMBER'])
            def HEX(self, tok):
                tok.type = 'NUMBER'
                tok.value = int(tok.value, 16)
            @token(r'\+\+')
            def INC(self, tok):
                tok.type = None
        self.lexer = TestedLexer(self)

    def test_single(self):
        token, = self.doLex('abc')
        self.assertEqual((token.type, token.value), ('ID', 'abc'))

    def test_ignore_leading(self):
        token, = self.doLex('  abc')
        self.assertEqual((token.type, token.value), ('ID', 'abc'))

    def test_ignore_middle(self):
        tok1, tok2 = self.doLex('a bc')
        self.assertEqual((tok1.type, tok1.value), ('ID', 'a'))
        self.assertEqual((tok2.type, tok2.value), ('ID', 'bc'))

    def test_ignore_trailing(self):
        token, = self.doLex('abc  ')
        self.assertEqual((token.type, token.value), ('ID', 'abc'))

    def test_value(self):
        token, = self.doLex('42')
        self.assertEqual((token.type, token.value), ('NUMBER', 42))

    def test_forced_value_eof(self):
        token, = self.doLex('abc\n')
        self.assertEqual((token.type, token.value), ('ID', 'abc'))

    def test_forced_value(self):
        token, = self.doLex('0xf')
        self.assertEqual((token.type, token.value), ('NUMBER', 15))

    def test_ignore(self):
        tok1, tok2 = self.doLex('a++b')
        self.assertEqual((tok1.type, tok1.value), ('ID', 'a'))
        self.assertEqual((tok2.type, tok2.value), ('ID', 'b'))

    def test_tokenvalues(self):
        self.assertEqual(self.lexer.token_types(), set(['ID', 'NUMBER', 'INC']))


class ReLexerBasicTestCase(LexerBasicTestCaseMixin, ReLexerTestCase):
    pass


class PositionTestCaseMixin(object):
    def setUp(self):
        super().setUp()
        class TestedLexer(LexerUnderTestMixin, self.lexerClass):
            @staticmethod
            def ignore(char):
                return char in [' ', '\n']
            @token('[a-z]+')
            def identifier(self, tok):
                pass

        self.lexer = TestedLexer(testCase=self)

    def test_error_position(self):
        try:
            self.doLex('ab\ncd0aa')
        except LexerError as exc:
            self.assertEqual(exc.lineno, 2)
            self.assertEqual(exc.colno, 3)
        else:
            self.fail()

    def test_token_positions(self):
        tok1, tok2 = self.doLex('ab\n  int')
        self.assertEqual(tok1.position, LexerPosition(column=1, line=1))
        self.assertEqual(tok2.position, LexerPosition(column=3, line=2))


class ReLexerPositionTestCase(PositionTestCaseMixin, ReLexerTestCase):
    pass


class TokenTypeTestCaseMixin(object):
    def setUp(self):
        super().setUp()
        class TestedLexer(self.lexerClass):
            def __init__(self, testCase):
                self.testCase = testCase
                super().__init__()
            @token('[a-z]', types=['LETTER'])
            def letter(self, tok):
                self.testCase.assertTrue(tok.type is None)
            def new_token(self, tok):
                pass
          
        self.lexer = TestedLexer(self)

    def test_none(self):
        self.doLex('a')

    def test_funcname(self):
        self.assertFalse('letter' in self.lexer.token_types())

    def test_types(self):
        self.assertTrue('LETTER' in self.lexer.token_types())


class ReLexerTokenTypeTestCase(TokenTypeTestCaseMixin, ReLexerTestCase):
    pass


class LexerByteTestCaseMixin(object):
    def setUp(self):
        super().setUp()
        class TestedLexer(LexerUnderTestMixin, self.lexerClass):
            @token(b'[a-zA-Z]+')
            def ID(self, tok):
                pass
        self.lexer = TestedLexer(self)

    def test_byte_regex_gives_byte_token_value(self):
        tok, = self.doLex(b'foo')
        self.assertTrue(isinstance(tok.value, bytes))


class ReLexerByteTestCase(LexerByteTestCaseMixin, ReLexerTestCase):
    pass


class LexerUnicodeTestCaseMixin(object):
    def setUp(self):
        super().setUp()
        class TestedLexer(LexerUnderTestMixin, self.lexerClass):
            @token('[a-zA-Z]+')
            def ID(self, tok):
                pass
        self.lexer = TestedLexer(self)

    def test_unicode_regex_gives_unicode_token_value(self):
        tok, = self.doLex('foo')
        self.assertTrue(isinstance(tok.value, str))


class ReLexerUnicodeTestCase(LexerUnicodeTestCaseMixin, ReLexerTestCase):
    pass


class LexerConsumerTestCaseMixin(object):
    def setUp(self):
        super().setUp()
        class TestedLexer(LexerUnderTestMixin, self.lexerClass):
            @token('[a-zA-Z0-9]+')
            def ID(self, tok):
                pass
            @token('"')
            def STR(self, tok):
                class StringBuilder(object):
                    def __init__(self):
                        self.value = io.StringIO()
                        self.state = 0
                    def feed(self, char):
                        if self.state == 0:
                            if char == '\\':
                                self.state = 1
                            elif char == '"':
                                return 'STR', self.value.getvalue()
                            else:
                                self.value.write(char)
                        elif self.state == 1:
                            self.value.write(char)
                            self.state = 0
                self.set_consumer(StringBuilder())
        self.lexer = TestedLexer(self)

    def test_string(self):
        tok1, tok2, tok3 = self.doLex(r'ab"foo\"spam"eggs')
        self.assertEqual((tok1.type, tok1.value), ('ID', 'ab'))
        self.assertEqual((tok2.type, tok2.value), ('STR', 'foo"spam'))
        self.assertEqual((tok3.type, tok3.value), ('ID', 'eggs'))


class ReLexerConsumerTestCase(LexerConsumerTestCaseMixin, ReLexerTestCase):
    pass


class LexerDuplicateTokenNameTestCaseMixin(object):
    def test_dup(self):
        try:
            class TestedLexer(LexerUnderTestMixin, self.lexerClass):
                @token('a')
                def ID(self, tok):
                    pass
                @token('b')
                def ID(self, tok):
                    pass
        except TypeError:
            pass
        else:
            self.fail()


class ReLexerDuplicateTokenNameTestCase(LexerDuplicateTokenNameTestCaseMixin, ReLexerTestCase):
    pass


class LexerInheritanceTestCaseMixin(object):
    def setUp(self):
        super().setUp()
        class TestedLexer(LexerUnderTestMixin, self.lexerClass):
            @token('[0-9]')
            def digit(self, tok):
                pass

        class ChildLexer(TestedLexer):
            def digit(self, tok):
                tok.value = int(tok.value)

        self.lexer = ChildLexer(self)

    def test_inherit(self):
        token, = self.doLex('4')
        self.assertEqual((token.type, token.value), ('digit', 4))


class ReLexerInheritanceTestCase(LexerInheritanceTestCaseMixin, ReLexerTestCase):
    pass


class LexerUnterminatedTokenTestCaseMixin(object):
    def setUp(self):
        super().setUp()
        class TestedLexer(LexerUnderTestMixin, self.lexerClass):
            @token('abc')
            def ID(self, tok):
                pass
        self.lexer = TestedLexer(self)

    def test_simple(self):
        token, = self.doLex('abc')
        self.assertEqual((token.type, token.value), ('ID', 'abc'))

    def test_unterminated(self):
        try:
            self.doLex('ab')
        except LexerError:
            pass
        else:
            self.fail()


class ReLexerUnterminatedTokenTestCase(LexerUnterminatedTokenTestCaseMixin, ReLexerTestCase):
    pass


class LexerLengthTestCaseMixin(object):
    def setUp(self):
        super().setUp()
        class TestedLexer(LexerUnderTestMixin, self.lexerClass):
            @token('<|=')
            def LT(self, tok):
                pass
            @token('<=')
            def LTE(self, tok):
                pass
        self.lexer = TestedLexer(self)

    def test_longest(self):
        token, = self.doLex('<=')
        self.assertEqual((token.type, token.value), ('LTE', '<='))


class ReLexerLengthTestCase(LexerLengthTestCaseMixin, ReLexerTestCase):
    pass


class LexerPriorityTestCaseMixin(object):
    def setUp(self):
        super().setUp()
        class TestedLexer(LexerUnderTestMixin, self.lexerClass):
            @token('a|b')
            def A(self, tok):
                pass
            @token('b|c')
            def B(self, tok):
                pass
        self.lexer = TestedLexer(self)

    def test_priority(self):
        token, = self.doLex('b')
        self.assertEqual((token.type, token.value), ('A', 'b'))


class ReLexerPriorityTestCase(LexerPriorityTestCaseMixin, ReLexerTestCase):
    pass


class LexerEOFTestCaseMixin(object):
    def setUp(self):
        super().setUp()
        class TestedLexer(LexerUnderTestMixin, self.lexerClass):
            @token(r'[0-9]+')
            def NUMBER(self, tok):
                tok.value = int(tok.value)
            @token(r'\n')
            def EOL(self, tok):
                tok.type = EOF
        self.lexer = TestedLexer(self)

    def test_eol_is_eof(self):
        self.lexer.parse('42\n')
        self.assertTrue(isinstance(self.tokens, tuple))


class ReLexerEOFTestCase(LexerEOFTestCaseMixin, ReLexerTestCase):
    pass


if __name__ == '__main__':
    unittest.main()
